"""
FastAPI server for Temporal Key-Value Store.
Exposes REST API endpoints for PUT and GET operations.
"""

from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, Annotated

import typer
import uvicorn
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from pydantic import BaseModel

from src.service import SQLiteStore

if TYPE_CHECKING:
    from collections.abc import AsyncIterator


class PutModel(BaseModel):
    """Request and response model for PUT operations."""

    key: str
    value: str
    timestamp: int


class GetModel(BaseModel):
    """Response model for successful GET operations."""

    value: str


@asynccontextmanager
async def lifespan(app: FastAPI) -> AsyncIterator[None]:  # noqa: RUF029
    """Manage database lifecycle with app state."""
    # Database lives only during app lifetime
    app.state.db = SQLiteStore()
    yield
    # Cleanup when app shuts down
    app.state.db.close()


def get_database(request: Request) -> SQLiteStore:
    """
    Get database from app state (shared across all requests).

    Parameters
    ----------
    request : Request
        The FastAPI request object

    Returns
    -------
    SQLiteStore
        The database instance from app state

    """
    return request.app.state.db


DatabaseDep = Annotated[SQLiteStore, Depends(get_database)]


app = FastAPI(
    title="Temporal Key-Value Store API",
    description="A key-value store that supports timestamp-based queries",
    version="1.0.0",
    lifespan=lifespan,
)


@app.put("/")  # NOTE: async ommitted as database operations are synchronous
def put_value(body: PutModel, db: DatabaseDep) -> PutModel:
    """
    Store a key-value pair with a timestamp.

    Parameters
    ----------
    body : PutModel
        The request body containing key, value, and timestamp
    db : SQLiteStore
        Database instance (injected dependency)

    Returns
    -------
    PutModel
        Response containing the stored key, value, and timestamp

    """
    db.put(body.key, body.value, body.timestamp)
    return PutModel(key=body.key, value=body.value, timestamp=body.timestamp)


@app.get("/{key}")
def get_value(key: str, timestamp: int, db: DatabaseDep) -> GetModel:
    """
    Retrieve a value for a key at a specific timestamp.

    Parameters
    ----------
    key : str
        The key to retrieve (path parameter)
    timestamp : int
        Timestamp to query (query parameter)
    db : SQLiteStore
        Database instance (injected dependency)

    Returns
    -------
    GetModel
        Response containing the value associated with the largest timestamp
        less than or equal to the provided timestamp

    Raises
    ------
    HTTPException
        If no value is found for the key at the specified timestamp.

    """
    value = db.get(key, timestamp)
    if value is None:
        raise HTTPException(
            status_code=404,
            detail=f"No value found for key '{key}' at timestamp {timestamp}",
        )
    return GetModel(value=value)


@app.exception_handler(Exception)
def fallback_exception_handler(_: Request, exc: Exception) -> JSONResponse:
    """
    Fallback handler for all unhandled exceptions.

    Parameters
    ----------
    exc : Exception
        The exception that was raised

    Returns
    -------
    JSONResponse
        JSON response with error message and status code 500

    """
    return JSONResponse(
        status_code=500,
        content={"message": "Internal server error", "detail": str(exc)},
    )


app_wrapper = typer.Typer(name="tkv-web", help="Run the FastAPI web server")


@app_wrapper.command()
def run(
    host: Annotated[str, typer.Option(help="Host to bind to")] = "127.0.0.1",
    port: Annotated[int, typer.Option(help="Port to bind to")] = 8000,
) -> None:
    """Run the FastAPI server."""
    uvicorn.run(app, host=host, port=port)


if __name__ == "__main__":
    run()
