"""
CLI interface for Temporal Key-Value Store.
Provides command-line interface for PUT and GET operations.
"""

from __future__ import annotations

from typing import Annotated

import typer
from rich import print as rprint

from src.service import SQLiteStore

app = typer.Typer(
    name="tkv",
    help="Temporal Key-Value Store CLI",
    rich_markup_mode="markdown",
)


@app.command(no_args_is_help=True)
def put(
    key: Annotated[str, typer.Argument(help="The key to store")],
    value: Annotated[str, typer.Argument(help="The value to store")],
    timestamp: Annotated[
        int, typer.Argument(help="Timestamp associated with key-value pair")
    ],
) -> None:
    """Store a key-value pair with timestamp."""
    with SQLiteStore() as store:
        store.put(key, value, timestamp)
        rprint(
            f"Stored [bold]{key}[/bold] = [green]{value}[/green] at timestamp"
            f" [blue]{timestamp}[/blue]"
        )


@app.command()
def get(
    key: Annotated[str, typer.Argument(help=" The key to retrieve the value for.")],
    timestamp: Annotated[int, typer.Argument(help="The timestamp to query.")],
) -> None:
    """
    Get value at key with largest timestamp less than or equal to the requested
    timestamp, if it exists.
    """
    with SQLiteStore() as store:
        value = store.get(key, timestamp)
        if value is None:
            rprint(
                f"No value found for key '[bold]{key}[/bold]' at any timestamp less"
                f" than or equal to [blue]{timestamp}[/blue]",
            )
        else:
            rprint(value)


if __name__ == "__main__":
    app()
