"""
Key-Value-Timestamp Store Implementations.
Multiple database backends for comparison.

Requirements:
1. Data persisted through process restarts (stored on disk)
2. PUT and GET APIs
3. Lookups consider key-timestamp combination (largest timestamp <= requested)
4. Defined outcomes for concurrent API calls
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Self

import rocksdict
from redislite import Redis  # type: ignore[reportMissingTypeStubs]
from sqlalchemy import create_engine
from sqlalchemy.event import listen
from sqlalchemy.orm import DeclarativeBase, Mapped, Session, mapped_column

if TYPE_CHECKING:
    import sqlite3

    from sqlalchemy.pool import ConnectionPoolEntry

DEFAULT_STORE_NAME = "tkv.db"

# =============================================================================
# ABSTRACT BASE CLASS FOR ALL IMPLEMENTATIONS
# =============================================================================


class KVTimestampStore[T](ABC):
    """Abstract base class for all key-value-timestamp store implementations."""

    @abstractmethod
    def put(self, key: str, value: T, timestamp: int) -> None:
        """Store key-value with timestamp."""

    @abstractmethod
    def get(self, key: str, timestamp: int) -> T | None:
        """
        Get value with largest timestamp less than or equal to requested
        timestamp.
        """

    @abstractmethod
    def close(self) -> None:
        """Close store and clean up resources."""

    def __enter__(self) -> Self:
        """
        Enter context manager, returning self.

        Returns
        -------
        Self
            The instance itself, allowing use in a `with` statement.

        """
        return self

    def __exit__(self, *_: object) -> None:
        """Exit context manager, cleaning up resources."""
        self.close()


# =============================================================================
# 1. SQLITE3 + SQLALCHEMY IMPLEMENTATION
# =============================================================================


class TKVBase(DeclarativeBase):
    """Base orm model for Key value timestamp store."""


class TKVRecord(TKVBase):
    """
    ORM model for a key-value-timestamp record.

    Attributes
    ----------
    key : str
        The key for the record.
    timestamp : int
        The timestamp for the record.
    value : str
        The value associated with the key and timestamp.

    Notes
    -----
    - key and timestamp uniquely identify a record and is hence is a composite
      primary key.

    """

    __tablename__ = "tkv_record"

    key: Mapped[str] = mapped_column(primary_key=True)
    timestamp: Mapped[int] = mapped_column(primary_key=True)
    value: Mapped[str]


class SQLiteStore(KVTimestampStore[str]):
    """
    SQLite-based key-value-timestamp store with WAL mode for concurrency and other
    performance optimizations.

    Attributes
    ----------
    engine : Engine
        SQLAlchemy engine for database connections

    Notes
    -----
    - Uses WAL mode for multiple readers and one writer
    - Uses SQLAlchemy ORM as database interface. Small overhead but safer (nice typing)
    and out-of-the-box optimizations.

    """

    def __init__(self, db_path: str = DEFAULT_STORE_NAME) -> None:
        """
        Connect to SQLite-based store at the specified path.

        Parameters
        ----------
        db_path : str, default = DEFAULT_STORE_NAME
            Path to SQLite database file.

        Notes
        -----
        - Connection does not run in async mode => incurs a performance penalty per user

        """
        # Create engine with connection pooling to reduce overhead of creating
        # and closing connections
        self.engine = create_engine(f"sqlite:///{db_path}")

        # Apply performance PRAGMAS on connect, including enabling WAL mode
        listen(self.engine, "connect", self.set_performance_pragmas)

        # Create table if it does not exist
        TKVRecord.metadata.create_all(self.engine)

    def put(self, key: str, value: str, timestamp: int) -> None:
        """
        Store key value pair with associated timestamp.

        Parameters
        ----------
        key : str
            The key to store
        value : str
            The value to store
        timestamp : int
            Timestamp associated with the key-value pair

        """
        with Session(self.engine) as session:
            record = TKVRecord(key=key, timestamp=timestamp, value=value)
            session.merge(record)  # Use merge for INSERT OR REPLACE behavior
            session.commit()

    def get(self, key: str, timestamp: int) -> str | None:
        """
        Get value at key with largest timestamp less than or equal to the requested
        timestamp, if it exists.

        Parameters
        ----------
        key : str
            The key to retrieve the value for
        timestamp : int
            The timestamp to query.

        Returns
        -------
        str | None
            The value associated with the key at the largest timestamp less than or
            equal to the requested timestamp, or None if no such value exists.

        Notes
        -----
        Value is retrieved using a single query with ORDER BY DESC and LIMIT 1.
        The PRIMARY KEY (key, timestamp) is sufficient for optimal performance:
            1. SQLite uses SeekLE operation to traverse the B+ tree defined by the
                composite primary key
            2. SeekLE finds the LARGEST timestamp <= target in O(log n) time
            3. Returns that record immediately (no sequential scanning)

        """
        with Session(self.engine) as session:
            return (
                session.query(TKVRecord.value)
                .filter(TKVRecord.key == key, TKVRecord.timestamp <= timestamp)
                .order_by(TKVRecord.timestamp.desc())
                .limit(1)
                .scalar()
            )

    def close(self) -> None:
        """Close SQLAlchemy engine and all connections."""
        self.engine.dispose()

    @staticmethod
    def set_performance_pragmas(
        conn: sqlite3.Connection, _: ConnectionPoolEntry | None = None
    ) -> None:
        """
        Set PRAGMA settings for optimized performance on SQLite connection.

        Parameters
        ----------
        conn : sqlite3.Connection
            Connection to a SQLite database

        Notes
        -----
        Performance PRAGMA settings applied:

        - synchronous=NORMAL : Forces data to disk only when absolutely necessary
            (at WAL file checkpoints and transaction commits) instead of after every
            write operation. Provides ~2x performance improvement while maintaining
            crash safety.

        - cache_size=10000 : Sets page cache to 10,000 pages. ~10-40MB depending on page
            size. Keeps frequently accessed B-tree index pages in memory, thereby
            improving timestamp range queries.

        - temp_store=memory : Forces temporary tables, indexes, and intermediate query
            results to be stored in RAM instead of disk. Eliminates I/O overhead for
            timestamp range queries.

        - mmap_size=268435456 : Treats first 256MB of database file as if it were
            regular memory rather than a file. When you read data, it comes directly
            from this "memory" without copying. OS automatically handles loading file
            pages into RAM when needed. Provides 2-10x improvement for read-heavy
            workloads.

        """
        conn.execute("PRAGMA journal_mode=WAL")
        conn.execute("PRAGMA synchronous=NORMAL")
        conn.execute("PRAGMA cache_size=10000")
        conn.execute("PRAGMA temp_store=memory")
        conn.execute("PRAGMA mmap_size=268435456")


# NOTE: below implentations are mainly here for the sake of research and may be
# unstable.

# =============================================================================
# 2. REDIS IMPLEMENTATION
# =============================================================================


class RedisStore(KVTimestampStore[str]):
    """
    Redis-based temporal key-value store using dual Hash+SortedSet architecture.

    Attributes
    ----------
    redis : Redis
        Redis client with persistence and performance optimizations

    Notes
    -----
    The dual structure approach solves the fundamental tension between fast
    value access and efficient temporal queries. Redis Streams cannot overwrite
    existing timestamps, making them unsuitable for mutable temporal data.
    Sorted sets alone fail because Redis member uniqueness means storing
    the same value at different timestamps overwrites the earlier entry.

    """

    def __init__(self, db_path: str = DEFAULT_STORE_NAME) -> None:
        """
        Initialize Redis with optimized configuration for temporal workloads.

        Parameters
        ----------
        db_path : str, default = DEFAULT_STORE_NAME
            Path to Redis database file.

        Notes
        -----
        Configuration optimizes for temporal access patterns: frequent writes
        with occasional range queries. AOF provides write durability while RDB
        enables fast restart recovery. Ziplist encoding reduces memory overhead
        for small timestamp collections common in temporal workloads.

        """
        self.redis = Redis(
            dbfilename=db_path,
            decode_responses=True,
            serverconfig={
                # === PERSISTENCE CONFIGURATION ===
                # Enable AOF (Append Only File) for durability
                "appendonly": "yes",
                # Sync AOF to disk every second (balance safety/performance)
                "appendfsync": "everysec",
                # Create RDB snapshots for backup
                "save": (
                    "900 1 300 10 60 10000"
                ),  # Save if: 1 change in 15min OR 10 in 5min OR 10k in 1min
                # AOF rewrite when file doubles in size
                "auto-aof-rewrite-percentage": "100",
                "auto-aof-rewrite-min-size": "64mb",
                # Stop writes if background save fails
                "stop-writes-on-bgsave-error": "yes",
                # Checksum RDB files for corruption detection
                "rdbchecksum": "yes",
                # === MEMORY MANAGEMENT ===
                # Set memory limit to prevent unlimited growth
                "maxmemory": "256mb",
                # Use LRU eviction when memory limit reached
                "maxmemory-policy": "allkeys-lru",
                # === DATA STRUCTURE OPTIMIZATIONS ===
                # Optimize hashes for our values:key usage pattern
                "hash-max-ziplist-entries": "1024",  # Keep small hashes compressed
                "hash-max-ziplist-value": "64",  # Optimize for timestamp keys
                # Optimize sorted sets for our timestamps:key usage pattern
                "zset-max-ziplist-entries": "256",  # Keep small sorted sets compressed
                "zset-max-ziplist-value": "64",  # Optimize for timestamp members
                # === PERFORMANCE TUNING ===
                # Client connection timeout
                "timeout": "300",
                # Keep TCP connections alive
                "tcp-keepalive": "60",
                # Background task frequency
                "hz": "10",
                # Disable expensive slowlog for better performance
                "slowlog-max-len": "0",
            },
        )

    def put(self, key: str, value: str, timestamp: int) -> None:
        """
        Store key-value pair with timestamp using atomic dual structure update.

        Parameters
        ----------
        key : str
            The key to store
        value : str
            The value to store
        timestamp : int
            Timestamp associated with the key-value pair

        """
        # Atomic dual structure update via Lua script
        # HSET stores timestamp→value mapping in hash
        # ZADD stores timestamp as both member and score in sorted set
        self.redis.eval(
            """
            redis.call('HSET', KEYS[1], ARGV[1], ARGV[2])
            redis.call('ZADD', KEYS[2], ARGV[1], ARGV[1])
            return 'OK'
            """,
            2,  # Number of keys passed to Lua script
            f"values:{key}",  # KEYS[1] - name of hash storing timestamp→value mappings
            f"timestamps:{key}",  # KEYS[2] - name of sorted set for timestamp ordering
            str(
                timestamp
            ),  # ARGV[1] - timestamp used for hash key and sorted set member/value
            value,  # ARGV[2] - value stored in hash at timestamp field
        )

    def get(self, key: str, timestamp: int) -> str | None:
        """
        Find value at the largest timestamp <= requested timestamp.

        Parameters
        ----------
        key : str
            The key to retrieve the value for
        timestamp : int
            The timestamp to query for temporal lookup

        Returns
        -------
        str | None
            Value at largest timestamp ≤ requested, or None if no such value exists

        """
        # Two-step atomic temporal lookup via Lua script
        # 1. ZREVRANGEBYSCORE finds largest timestamp <= target (binary search O(log N))
        # 2. HGET retrieves corresponding value from hash (O(1))
        # Returns nil if no valid timestamp exists
        return self.redis.eval(  # type: ignore[reportReturnType]
            """
            local members = redis.call('ZREVRANGEBYSCORE', KEYS[1], ARGV[1], '-inf',
                                       'LIMIT', 0, 1)
            if #members == 0 then
                return nil
            end
            return redis.call('HGET', KEYS[2], members[1])
            """,
            2,  # Two Redis keys required for dual structure lookup
            f"timestamps:{key}",  # KEYS[1] - name of sorted set for timestamps
            f"values:{key}",  # KEYS[2] - name of hash for values
            str(timestamp),  # ARGV[1] - upper bound for ZREVRANGEBYSCORE query
        )

    def close(self) -> None:
        """Close Redis connection and ensure data is persisted."""
        # Force save to disk before closing
        self.redis.shutdown(save=True)  # type: ignore[reportUnknownMemberType]


# =============================================================================
# 6. ROCKSDB IMPLEMENTATION
# =============================================================================


class RocksDBStore[T](KVTimestampStore[T]):
    """

    RocksDB-based key-value-timestamp store using LSM trees with optimizations.

    Attributes
    ----------
    db_path : str
        Path to RocksDB database directory

    Notes
    -----
    - Uses composite keys for lexicographic ordering of key:timestamp pairs
    - Optimized for point lookups and write performance
    - Implements O(log n) retrieval using smart reverse iteration with stop condition
    - Designed for high concurrency and low write amplification
    - Uses RocksDB's advanced features for optimal performance

    """

    def __init__(self, db_path: str = DEFAULT_STORE_NAME) -> None:
        """
        Initialize RocksDB database with optimizations.

        Parameters
        ----------
        db_path : str, default = DEFAULT_STORE_NAME
            Path to RocksDB database directory.

        """
        options = rocksdict.Options()

        # Write optimization (30-50% improvement)
        options.set_write_buffer_size(64 * 1024 * 1024)  # 64MB
        options.set_max_write_buffer_number(3)  # Non-blocking writes during flush
        options.set_min_write_buffer_number_to_merge(1)  # Immediate merge

        # Compaction optimization (15-25% write amplification reduction)
        options.set_level_compaction_dynamic_level_bytes(True)
        options.set_level_zero_file_num_compaction_trigger(4)  # Optimal trigger point
        options.set_level_zero_slowdown_writes_trigger(20)  # Prevent write stalls
        options.set_level_zero_stop_writes_trigger(36)  # Emergency brake
        options.set_max_bytes_for_level_base(256 * 1024 * 1024)  # 256MB base level
        options.set_max_bytes_for_level_multiplier(10)  # 10x fanout ratio

        # I/O optimization
        options.set_use_direct_reads(False)  # Better for small datasets
        options.set_use_direct_io_for_flush_and_compaction(False)
        options.set_allow_mmap_reads(True)  # Memory-mapped reads

        # Optimize for point lookups
        options.optimize_for_point_lookup(256 * 1024 * 1024)  # 256MB block cache

        # Maximum performance configuration
        self.db = rocksdict.Rdict(db_path, options)

    def put(self, key: str, value: T, timestamp: int) -> None:
        """
        Store the key-value pair with associated timestamp using a composite key format.

        Parameters
        ----------
        key : str
            The key to store
        value : T
            The value to store
        timestamp : int
            Timestamp associated with the key-value pair

        Notes
        -----
        rocksdict automatically handles serialization using pickle for all
        Python objects. No manual serialization needed.

        """
        composite_key = self.make_composite_key(key, timestamp)
        self.db[composite_key] = value

    def get(self, key: str, timestamp: int) -> T | None:
        """
        Get value at key with largest timestamp less than or equal to the requested
        timestamp, if it exists.

        Parameters
        ----------
        key : str
            The key to retrieve the value for
        timestamp : int
            The timestamp at which to retrieve the value

        Returns
        -------
        T | None
            The value associated with the key at the largest timestamp less than or
            equal to the requested timestamp, or None if no such value exists.

        Notes
        -----
        Uses reverse iteration of the store. The provided key defines the starting
        point which is found in O(log n) time using RocksDB's LSM tree structure.
        Because the iterator itself is reverse lexicographically ordered, this starting
        point must be the value we are looking for, if it exists, so we can return early
        without doing any scanning.

        Raises
        ------
        TypeError
            If the composite key is not a string.

        """
        target_key = self.make_composite_key(key, timestamp)

        for db_key, value in self.db.items(backwards=True, from_key=target_key):
            if not isinstance(db_key, str):
                raise TypeError
            if db_key.startswith(f"{key}:") and db_key <= target_key:
                return value  # rocksdict auto-deserializes with correct type
            break

        return None

    def close(self) -> None:
        """Close RocksDB database."""
        self.db.close()

    @staticmethod
    def make_composite_key(key: str, timestamp: int) -> str:
        """
        Create composite key for lexicographic ordering.

        Parameters
        ----------
        key : str
            The key to store
        timestamp : int
            The timestamp associated with the key

        Returns
        -------
        str
            Composite key in the format "key:timestamp" with proper ordering for
            negative numbers

        """
        # Add bias to handle negative numbers correctly in lexicographic ordering
        # 2^63 = 9223372036854775808, ensures all int64 values become positive
        biased_timestamp = timestamp + 9223372036854775808
        return f"{key}:{biased_timestamp:020d}"

    @staticmethod
    def split_composite_key(composite_key: str) -> tuple[str, str]:
        """
        Split composite into key and timestamp.

        Parameters
        ----------
        composite_key : str
            Composite key in the format "key:timestamp"

        Returns
        -------
        tuple[str, str]
            The key and timestamp extracted from the composite key

        Notes
        -----
        This method assumes the composite key is well-formed and contains a single
        colon separating the key and timestamp.

        """
        key, timestamp = composite_key.rsplit(":", 1)
        return key, timestamp
